#import <objc/runtime.h>
#import "client.h"
#import "base64.h"
#import "resource.h"
#import "context.h"

NSString *const APIURL = @"https://api.stylegeist.com";


@interface SGClient ()
{
	NSMutableDictionary *_HTTPMethods;
	BOOL _error;
}

@property CFMutableDictionaryRef connectionToInfoMapping;
@property NSString *boundary;

@end


@implementation SGClient

+ (id)hiddenAlloc
{
	return [super alloc];
}

+ (id)alloc
{
	NSLog(@"SGClient: use +sharedInstance instead of +alloc");
	return nil;
}

+ (SGClient *)sharedInstance
{
	static SGClient *c = nil;
	if (!c) {
		c = [[[self class] hiddenAlloc] init];
		c.connectionToInfoMapping = CFDictionaryCreateMutable(kCFAllocatorDefault,
			0,
			&kCFTypeDictionaryKeyCallBacks,
			&kCFTypeDictionaryValueCallBacks);
	}

	return c;
}

static const char *getPropertyType(objc_property_t property)
{
    const char *attributes = property_getAttributes(property);
    char buffer[1 + strlen(attributes)];
    strcpy(buffer, attributes);
    char *state = buffer, *attribute;
    while ((attribute = strsep(&state, ",")) != NULL) {
        if (attribute[0] == 'T') {
            return (const char *)[[NSData dataWithBytes:(attribute + 3)
												 length:strlen(attribute) - 4
								   ] bytes];
        }
    }
    return "@";
}

- (NSString *)setBoundary
{
	_boundary = @"----BOUNDARY";
	return _boundary;
}

- (void)save:(SGResource *)rsc
{
	NSMutableURLRequest *r = [NSMutableURLRequest requestWithURL:rsc.url
													 cachePolicy:NSURLRequestUseProtocolCachePolicy
												 timeoutInterval:60.0];

	if ([rsc isUpdated]) {
		r.HTTPMethod = @"PUT";
	} else {
		r.HTTPMethod = @"POST";
	}

	[self setBoundary];
	NSString *ct = [NSString stringWithFormat:@"multipart/form-data; boundary=%@", _boundary];
	[r setValue:ct forHTTPHeaderField:@"Content-Type"];
	[r setValue:@"application/json" forHTTPHeaderField:@"Accept"];
	NSMutableData *data = [[NSMutableData alloc] init];

	NSString *as = [NSString stringWithFormat:@"%@:%@", [self username], [self password]];
	NSData *ad = [as dataUsingEncoding:NSUTF8StringEncoding];
	NSString *av = [NSString stringWithFormat:@"Basic %@", [ad base64EncodedString]];
	[r setValue:av forHTTPHeaderField:@"Authorization"];

	[data appendData:[[NSString stringWithFormat:@"--%@\r\n", _boundary] dataUsingEncoding:NSUTF8StringEncoding]];
	NSString *contentDisposition = [NSString stringWithFormat:@"Content-Disposition: form-data; name=utf8;\r\n\r\n"];
	[data appendData:[contentDisposition dataUsingEncoding:NSUTF8StringEncoding]];
	[data appendData:[@"✓\r\n" dataUsingEncoding:NSUTF8StringEncoding]];

	unsigned int outCount, i;
	objc_property_t *properties = class_copyPropertyList([rsc class], &outCount);

	for (i = 0; i < outCount; i++) {
		objc_property_t p = properties[i];
		NSString *pname = [NSString stringWithCString:property_getName(p) encoding:NSASCIIStringEncoding];

		NSLog(@"— Excuse me, sir, are we updating %@?", pname);

		if ([rsc valueForKey:pname] && ![pname isEqualToString:@"original"]) {
			NSLog(@"— Yes, indeed!");
			// if it's a copy, optimistically update the original

			if ([rsc isUpdated]) {
				[[rsc valueForKey:@"original"] setValue:[rsc valueForKey:pname] forKey:pname];
			}

			const char *s = getPropertyType(p);
			NSString *type = [NSString stringWithCString:s encoding:NSASCIIStringEncoding];
			if ([type isEqualToString:@"NSData"]) {
				NSData *b = [[NSString stringWithFormat:@"--%@\r\n", _boundary] dataUsingEncoding:NSUTF8StringEncoding];
				NSData *cd = [NSString stringWithFormat:@"Content-Disposition: form-data; name=\"%@[%@]\"; filename=\"image.jpg\"\r\n", [rsc valueForKey:@"remoteEntityName"], pname];
				NSData *ct1 = [@"Content-Type: image/jpeg\r\n" dataUsingEncoding:NSUTF8StringEncoding];
				NSData *ct2 = [@"Content-Type: application/octet-stream\r\n\r\n" dataUsingEncoding:NSUTF8StringEncoding];
				NSData *cr = [@"\r\n" dataUsingEncoding:NSUTF8StringEncoding];

				[data appendData:b];
				[data appendData:cd];
				[data appendData:ct1];
				[data appendData:ct2];
				[data appendData:[self valueForKey:pname]];
				[data appendData:cr];
			} else if ([type isEqualToString:@"NSString"]) {
				[data appendData:[[NSString stringWithFormat:@"--%@\r\n", _boundary] dataUsingEncoding:NSUTF8StringEncoding]];
				NSString *contentDisposition = [NSString stringWithFormat:@"Content-Disposition: form-data; name=\"%@[%@]\"\r\n\r\n", [rsc valueForKey:@"remoteEntityName"], pname];
				[data appendData:[contentDisposition dataUsingEncoding:NSUTF8StringEncoding]];
				[data appendData:[(NSString *)[self valueForKey:pname] dataUsingEncoding:NSUTF8StringEncoding]];
				[data appendData:[@"\r\n" dataUsingEncoding:NSUTF8StringEncoding]];
			} else if ([type isEqualToString:@"NSNumber"]) {
				[data appendData:[[NSString stringWithFormat:@"--%@\r\n", _boundary] dataUsingEncoding:NSUTF8StringEncoding]];
				NSString *contentDisposition = [NSString stringWithFormat:@"Content-Disposition: form-data; name=\"%@[%@]\"\r\n\r\n", [rsc valueForKey:@"remoteEntityName"], pname];
				[data appendData:[contentDisposition dataUsingEncoding:NSUTF8StringEncoding]];
				NSNumber *number = [self valueForKey:pname];
				NSString *stringedNumber = [number stringValue];
				[data appendData:[stringedNumber dataUsingEncoding:NSUTF8StringEncoding]];
				[data appendData:[@"\r\n" dataUsingEncoding:NSUTF8StringEncoding]];
			}
		} else {
			NSLog(@"— No, sorry.");
		}
	}

	[data appendData:[[NSString stringWithFormat:@"--%@--\r\n", _boundary] dataUsingEncoding:NSUTF8StringEncoding]];

	NSString *dataString = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
	NSLog(@"%@", dataString);

	[r setHTTPBody:data];
	[r setValue:[NSString stringWithFormat:@"%d", [data length]] forHTTPHeaderField:@"Content-Length"];

	NSURLConnection *c = [[NSURLConnection alloc] initWithRequest:r delegate:self startImmediately:YES];

	NSMutableDictionary *info = [@{
								 @"receivedData" : [NSMutableData data],
								 @"HTTPMethod" : r.HTTPMethod,
								 @"resource" : rsc
								 } mutableCopy];

    CFDictionaryAddValue(_connectionToInfoMapping, (__bridge const void *)(c), (__bridge const void *)(info));
}

- (void)load:(SGResource *)rsc
{
	NSMutableURLRequest *r = [NSMutableURLRequest requestWithURL:rsc.url
													 cachePolicy:NSURLRequestUseProtocolCachePolicy
												 timeoutInterval:60];

	[r setValue:@"application/json" forHTTPHeaderField:@"Accept"];

	NSString *as = [NSString stringWithFormat:@"%@:%@", _username, _password];
	NSData *ad = [as dataUsingEncoding:NSUTF8StringEncoding];
	NSString *av = [NSString stringWithFormat:@"Basic %@", [ad base64EncodedString]];
	[r setValue:av forHTTPHeaderField:@"Authorization"];

	NSURLConnection *c = [[NSURLConnection alloc] initWithRequest:r
														 delegate:self
												 startImmediately:YES];

	NSMutableDictionary *info = [@{@"receivedData" : [NSMutableData data],
								 @"HTTPMethod" : r.HTTPMethod,
								 @"resource" : rsc
								 } mutableCopy];

	CFDictionaryAddValue(_connectionToInfoMapping,
						 (__bridge const void *)(c),
						 (__bridge const void *)(info)
						 );
}

- (void)connection:(NSURLConnection *)c didReceiveResponse:(NSHTTPURLResponse *)resp
{
	NSMutableDictionary *info = CFDictionaryGetValue(_connectionToInfoMapping, (__bridge const void *)(c));
	// reset the received data for the current connection => is this necessary with the dictionary?
	[[info objectForKey:@"receivedData"] setLength:0];
	NSNumber *code = [NSNumber numberWithInt:[resp statusCode]];
	[info setValue:code forKey:@"statusCode"];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
	NSMutableDictionary *info = CFDictionaryGetValue(_connectionToInfoMapping, (__bridge const void *)(connection));
	[[info objectForKey:@"receivedData"] appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
	NSMutableDictionary *info = CFDictionaryGetValue(_connectionToInfoMapping, (__bridge const void *)(connection));
	SGResource *rsc = [info valueForKey:@"resource"];
	if ([rsc.context.delegate respondsToSelector:@selector(connection:didFailForResource:inContext:withError:)]) {
		[rsc.context.delegate connection:connection didFailForResource:rsc inContext:rsc.context withError:error];
	}
}

- (void)logReceivedData:(NSData *)data
{
	NSLog(@"Succeeded! Received %d bytes of data", [data length]);
	NSLog(@"%@", [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding]);
}

- (void)connectionDidFinishLoading:(NSURLConnection *)c {
	NSMutableDictionary *i = CFDictionaryGetValue(_connectionToInfoMapping, (__bridge const void *)(c));
	NSData *d = [i valueForKey:@"receivedData"];
	NSString *m = [i valueForKey:@"HTTPMethod"];
	SGResource *r = [i valueForKey:@"resource"];
	int code = [[i valueForKey:@"statusCode"] intValue];

	if ([m isEqualToString:@"GET"]) {
		NSJSONSerialization *json = [NSJSONSerialization JSONObjectWithData:d options:0 error:NULL];
		[r mapFromJSONSerialization:json];
	}

	if (code == 401) {
		if ([r.context.delegate respondsToSelector:@selector(connection:didFailForResource:inContext:withError:)]) {
			[r.context.delegate connection:c didFailForResource:r inContext:r.context withError:[NSError errorWithDomain:@"SGSDK" code:code userInfo:i]];
		}
		return;
	} 

	if ([r.context.delegate respondsToSelector:@selector(connection:didSucceedForResource:inContext:)]) {
		[r.context.delegate connection:c didSucceedForResource:r inContext:r.context];
	}
}

- (SGContext *)contextWithDelegate:(id)d
{
	SGContext *c = [[SGContext alloc] init];
	c.client = self;
	c.delegate = d;
	return c;
}

- (SGContext *)context
{
	return [self contextWithDelegate:nil];
}

@end
