#import <objc/runtime.h>
#import "resource.h"
#import "desc.m"


@interface SGResource ()

@end


@implementation SGResource

- (id)initWithContext:(SGContext *)c remoteEntityName:(NSString *)n
{
	self = [super init];

	if (self) {
		_context = c;
		_remoteEntityName = n;
	}

	return self;
}

- (id)initWithContext:(SGContext *)c
{
	return [self initWithContext:c remoteEntityName:nil];
}

- (void)mapFromJSONSerialization:(NSJSONSerialization *)serialization
{
	if ([serialization valueForKeyPath:@"links.edit"]) {
		_editable = YES;
	}
}

- (BOOL)isUpdated
{
	return _original != nil;
}

- (SGResource *)updatedCopy
{
	SGResource *r = [[[self class] alloc] init];
	r.url = _url;
	r.remoteEntityName = _remoteEntityName;
	r.original = self;
	
	return r;
}


- (SGResource *)next
{
	NSLog(@"%i\n", [self respondsToSelector:@selector(nextURL)]);
	NSLog(@"%i\n", [[self valueForKey:@"nextURL"] absoluteString].length > 0);

	BOOL sel = [self respondsToSelector:@selector(nextURL)];
	BOOL empty = [[self valueForKey:@"nextURL"] absoluteString].length == 0;
	
	if (!sel || empty) {
		return nil;
	}

	SGResource *c = [[[self class] alloc] init];
	c.url = [self valueForKey:@"nextURL"];
	c.context = self.context;
	return c;
}

@end


@implementation SGStyle
{
	NSURL *_commentsURL;
	NSURL *_ratingsURL;
}

- (id)initWithContext:(SGContext *)c
{
	return [self initWithContext:c remoteEntityName:@"style"];
}

- (void)mapFromJSONSerialization:(NSJSONSerialization *)j
{
	[super mapFromJSONSerialization:j];
	_pictureURL = [NSURL URLWithString:[j valueForKeyPath:@"links.picture"]];
	_thumbnailURL = [NSURL URLWithString:[j valueForKeyPath:@"links.thumbnail"]];
	_commentsURL = [NSURL URLWithString:[j valueForKeyPath:@"links.comments"]];
	_ratingsURL = [NSURL URLWithString:[j valueForKeyPath:@"links.ratings"]];
	_user = [[SGUser alloc] init];
	[_user mapFromJSONSerialization:[j valueForKey:@"user"]];

	if ([j valueForKey:@"next"]) {
		_next = [[SGStyle alloc] init];
		[_next mapFromJSONSerialization:[j valueForKey:@"next"]];
	}

	if ([j valueForKey:@"prev"]) {
		_prev = [[SGStyle alloc] init];
		[_prev mapFromJSONSerialization:[j valueForKey:@"prev"]];
	}
}

@end


@implementation SGSession

- (id)initWithContext:(SGContext *)c
{
	return [self initWithContext:c remoteEntityName:@"session"];
}

- (void)mapFromJSONSerialization:(NSJSONSerialization *)serialization
{
	_user = [[SGUser alloc] init];
	_user.url = [NSURL URLWithString:[serialization valueForKeyPath:@"user.self"]];
}

@end


@implementation SGUser

- (id)initWithContext:(SGContext *)c
{
	return [self initWithContext:c remoteEntityName:@"user"];
}

- (void)mapFromJSONSerialization:(NSJSONSerialization *)serialization
{
	[super mapFromJSONSerialization:serialization];

	if ([serialization valueForKey:@"tagline"]) {
		_tagline = [serialization valueForKey:@"tagline"];
	}
	_username = [serialization valueForKey:@"username"];
	_email = [serialization valueForKey:@"email"];
	_pictureURL = [NSURL URLWithString:[serialization valueForKeyPath:@"links.picture"]];

	_styles = [[SGStylesCollection alloc] init];
	[_styles mapFromJSONSerialization:[serialization valueForKey:@"styles"]];
}

@end


@implementation SGComment

- (id)initWithContext:(SGContext *)c
{
	return [self initWithContext:c remoteEntityName:@"comment"];
}

- (void)mapFromJSONSerialization:(NSJSONSerialization *)serialization
{
	[super mapFromJSONSerialization:serialization];

	_content = [serialization valueForKey:@"content"];
	_user = [[SGUser alloc] init];
	[_user mapFromJSONSerialization:[serialization valueForKey:@"user"]];
}

@end


@implementation SGRating

- (id)initWithContext:(SGContext *)c
{
	return [self initWithContext:c remoteEntityName:@"rating"];
}

- (void)mapFromJSONSerialization:(NSJSONSerialization *)serialization
{
	[super mapFromJSONSerialization:serialization];

	_average = [serialization valueForKey:@"average"];
	_own = [serialization valueForKey:@"own"];
}

@end


@implementation SGStylesCollection

- (void)mapFromJSONSerialization:(NSJSONSerialization *)json
{
	_styles = [[NSMutableArray alloc] init];

	for (id j in [json valueForKey:@"styles"]) {
		SGStyle *s = [[SGStyle alloc] init];
		[s mapFromJSONSerialization:j];
		[_styles addObject:s];
	}
	_total = [json valueForKey:@"total"];
	_nextURL = [NSURL URLWithString:[json valueForKeyPath:@"links.next"]];
	NSLog(@"%@", _nextURL);
}

@end


@implementation SGUsersCollection

- (void)mapFromJSONSerialization:(NSJSONSerialization *)json
{
	_users = [[NSMutableArray alloc] init];
	for (id j in [json valueForKey:@"users"]) {
		SGUser *u = [[SGUser alloc] init];
		[u mapFromJSONSerialization:j];
		[_users addObject:u];
	}
	_total = [json valueForKey:@"total"];
	_nextURL = [NSURL URLWithString:[json valueForKeyPath:@"links.next"]];
}

@end


@implementation SGCommentsCollection

- (void)mapFromJSONSerialization:(NSJSONSerialization *)json
{
	_comments = [[NSMutableArray alloc] init];
	for (id j in [json valueForKey:@"comments"]) {
		SGComment *c = [[SGComment alloc] init];
		[c mapFromJSONSerialization:j];
		[_comments addObject:c];
	}
	_total = [json valueForKey:@"total"];
	_nextURL = [NSURL URLWithString:[json valueForKeyPath:@"links.next"]];
}

@end
